﻿using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.SceneManagement;
using UnityEngine.Serialization;
using UnityEngine.UI;

namespace AudioMate {
    /// <summary>
    /// AudioMate by dub
    /// -------------------------------------------------------
    /// Virt-a-Mate plugin for audio clip playback
    /// and audio clip organisation
    /// -------------------------------------------------------
    /// https://github.com/dub-a-licious/audio-mate
    /// </summary>
    public class AudioMateController : MVRScript
    {
        public static string GuessInitialReceivingNode(Atom atom)
        {
            switch (atom.category)
            {
                case "People":
                    return "HeadAudioSource";
                case "Sound":
                    switch (atom.type)
                    {
                        case "AudioSource":
                            return "AudioSource";
                        case "RhythmAudioSource":
                            return "RhythmSource";
                        case "AptSpeaker":
                            return "AptSpeaker_Import";
                        default:
                            return "None";
                    }
                default:
                    return "None";
            }
        }
        public class ContainingAtomNameChangedEventArgs { public string Before; public string After; }
        public class ContainingAtomNameChangedEvent : UnityEvent<ContainingAtomNameChangedEventArgs> {};
        public readonly ContainingAtomNameChangedEvent OnContainingAtomNameChanged = new ContainingAtomNameChangedEvent();

        private Scene _activeScene;

        public CollectionManager collections;
        [FormerlySerializedAs("audioMateClipLibrary")] public ClipLibrary clipLibrary;
        public UIManager ui;
        public new MVRPluginManager Manager => base.manager;
        public FileManager fileManager;
        public TriggerManager Triggers;

        public Atom receivingAtom;
        public JSONStorable receivingNode;

        public bool needsProvisionalClipProcessing;
        public bool uiInitialized;

        private bool _initializingUI;

        private string _missingReceiverNodeStoreId = "";
        private string _atomCreationId = null;

        // public AdjustJoints jawControl;
        // private JSONStorable _autoJawMouthMorph;

        private JSONStorableString _fileManagerJSON;
        public JSONStorableStringChooser CollectionsJSON;
        private JSONStorableAction _playClipActionJSON;
        private JSONStorableAction _queueClipActionJSON;
        private JSONStorableAction _playRandomClipActionJSON;
        private JSONStorableAction _queueRandomClipActionJSON;
        public JSONStorableStringChooser ReceivingAtomJSON;
        public JSONStorableStringChooser ReceivingNodeJSON;
        public JSONStorableStringChooser TriggerColliderChooserJSON;

        // Global Settings
        public JSONStorableBool DisplayHintsJSON;
        public JSONStorableBool VolumeControlEnabledJSON;
        public JSONStorableBool PitchControlEnabledJSON;

        public ClipCursor Cursor => clipLibrary.Cursor;
        public List<AudioMateClip> Clips => clipLibrary.Clips;
        public bool DisplayHints => DisplayHintsJSON.val;

        #region Initialization
        public override void Init()
        {
            Log("Initializing Main Controller");
            try
            {
                _activeScene = SceneManager.GetActiveScene();
                base.Init();
                Triggers = new TriggerManager(this);
                InitStorables();
                InitFileManager();
                StartCoroutine(DeferredInit());
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(Init)}: {e}");
            }
        }

        protected override void Awake()
        {
            SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
            base.Awake();
        }

        private void OnSceneLoaded()
        {
            if (this != null && collections != null && collections) containingAtom.RestoreFromLast(this);
        }

        private IEnumerator DeferredInit()
        {
            Log("Starting deferred initialization routine");
            if (collections == null)
            {
                while (_activeScene.isLoaded == false || SuperController.singleton.isLoading)
                {
                    Log("Waiting for scene to finish loading...");
                    yield return null;
                    if (this == null) yield break;
                }

                while (!uiInitialized)
                {
                    Log("Waiting for UI to be initialized...");
                    yield return new WaitForSeconds(.1f);
                    if (this == null) yield break;
                }
                InitPhysics();
                InitCollections();
                InitClipLibrary();
                clipLibrary.Bind();
            }

            Log("Trying to restore from last save...");
            containingAtom.RestoreFromLast(this);
        }

        public override void InitUI()
        {
            Log("Initializing UI");

            base.InitUI();
            try
            {
                if (UITransform == null) return;

                SuperController.singleton.StartCoroutine(InitUIDeferred());
            }
            catch (Exception exc)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(InitUI)}: {exc}");
            }
        }

        private IEnumerator InitUIDeferred()
        {
            if (ui != null ||
                _initializingUI) yield break;
            _initializingUI = true;
            yield return SuperController.singleton.StartCoroutine(UIFactory.LoadUIAssets());

            var scriptUI = UITransform.GetComponentInChildren<MVRScriptUI>();
            var scrollRect = scriptUI.fullWidthUIContent.transform.parent.parent.parent.GetComponent<ScrollRect>();

            if (scrollRect == null)
                SuperController.LogError(
                    "AudioMate: Error during UI initialization. Scroll rect not at the expected hierarchy position");
            else
            {
                scrollRect.vertical = false;
                scrollRect.horizontal = false;
                scrollRect.elasticity = 0;
                scrollRect.inertia = false;
                scrollRect.movementType = ScrollRect.MovementType.Clamped;
            }

            ui = UIManager.AddTo(scriptUI.fullWidthUIContent);
            ui.Init(this);
            var count = 0;
            while (collections == null || !collections.isInitialized)
            {
                // security break out to avoid endless loops
                if (++count > 200)
                {
                    SuperController.LogError(!collections.isInitialized
                        ? "AudioMate: Could not initialize collection manager."
                        : "AudioMate: Collection manager is null.");

                    yield break;
                }
                yield return new WaitForSeconds(.1f);
                if (this == null) yield break;
            }
            try
            {
                ui.Bind();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate: in InitUIDeferred - {e}");
            }
        }

        public void OnUIInitialized()
        {
            Log("Initialized UI");
            uiInitialized = true;
        }

        private void InitPhysics()
        {
            // jawControl = containingAtom.GetStorableByID("JawControl") as AdjustJoints;
            // _autoJawMouthMorph = containingAtom.GetStorableByID("AutoJawMouthMorph");
        }

        private void InitCollections()
        {
            if ((UnityEngine.Object) collections != (UnityEngine.Object) null) return;
            Log("Initializing Collections");
            try
            {
                collections = gameObject.AddComponent<CollectionManager>();
                collections.transform.SetParent(transform, false);

                collections.OnActiveCollectionSelected.RemoveAllListeners();
                collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);
                collections.OnActiveCollectionNameChanged.RemoveAllListeners();
                collections.OnActiveCollectionNameChanged.AddListener(OnActiveCollectionNameChanged);
                _playClipActionJSON = new JSONStorableAction(Storables.PlayClipAction,
                    () => collections.PlayClipAction());
                RegisterAction(_playClipActionJSON);
                _queueClipActionJSON = new JSONStorableAction(Storables.QueueClipAction,
                    () => collections.QueueClipAction());
                RegisterAction(_queueClipActionJSON);
                _playRandomClipActionJSON = new JSONStorableAction(Storables.PlayRandomClipAction,
                    () => collections.PlayRandomClipAction());
                RegisterAction(_playRandomClipActionJSON);
                _queueRandomClipActionJSON = new JSONStorableAction(Storables.QueueRandomClipAction,
                    () => collections.QueueRandomClipAction());
                RegisterAction(_queueRandomClipActionJSON);
                collections.Init(this);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(InitCollections)}: {e}");
            }
        }

        private void InitFileManager()
        {
            Log("Initializing File Manager");
            fileManager = gameObject.AddComponent<FileManager>();
            fileManager.Bind(_fileManagerJSON);
        }

        private void InitClipLibrary()
        {
            Log("Initializing Clip Library");
            if (clipLibrary == null) clipLibrary = gameObject.AddComponent<ClipLibrary>();
            clipLibrary.Init(this);
        }

        private void InitStorables()
        {
            Log("Initializing Storables");

            CollectionsJSON = new JSONStorableStringChooser(
                Storables.ActiveCollection,
                new List<string>(),
                null,
                "Active Collection")
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            RegisterStringChooser(CollectionsJSON);

            ReceivingAtomJSON = new JSONStorableStringChooser(Storables.ReceivingAtom, null,
                containingAtom.uid, "Receiver Atom", SetReceivingAtom)
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(ReceivingAtomJSON);

            ReceivingNodeJSON = new JSONStorableStringChooser(Storables.ReceivingNode, null,
                GuessInitialReceivingNode(containingAtom), "Receiver Node", SetReceivingNode)
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(ReceivingNodeJSON);

            TriggerColliderChooserJSON = new JSONStorableStringChooser(
                Storables.TriggerActions,
                Triggers.ColliderChoices,
                TriggerManager.DefaultCollider,
                "Trigger")
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(TriggerColliderChooserJSON);

            _fileManagerJSON = new JSONStorableString("File Manager", "")
            {
                isStorable = false,
                isRestorable = false
            };
            RegisterString(_fileManagerJSON);

            // Global Settings
            DisplayHintsJSON = new JSONStorableBool(
                Storables.DisplayHints,
                false);
            RegisterBool(DisplayHintsJSON);

            VolumeControlEnabledJSON = new JSONStorableBool(
                Storables.VolumeControlEnabled,
                false,
                val => collections.SetVolumeControl(val));
            RegisterBool(VolumeControlEnabledJSON);

            PitchControlEnabledJSON = new JSONStorableBool(
                Storables.PitchControlEnabled,
                false,
                val => collections.SetPitchControl(val));
            RegisterBool(PitchControlEnabledJSON);
            // Bind onAtomUIDRename event
            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
        }
        #endregion

        #region Lifecycle
        public void OnEnable()
        {
            if (collections == null) return;
            collections.EnableAll();
        }

        public void OnDisable()
        {
            if (collections != null) collections.DisableAll();
        }

        public void Start()
        {
            Log("Start");
        }

        public void OnDestroy()
        {
            Log("Destroying Plugin instance");
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            Triggers.OnDestroy();
            if (ui != null) Destroy(ui);
            if (collections == null) return;
            collections.OnDestroy();
            Destroy(collections);
        }

        public void Update()
        {
            if (!uiInitialized || collections == null) return;
            if (clipLibrary != null && needsProvisionalClipProcessing && clipLibrary.isInitialised)
            {
                needsProvisionalClipProcessing = false;
                clipLibrary.ProcessProvisionalClips();
            }
            FindMissingReceiverNode();
            collections.StartAutoPlayRoutine();
        }

        /**
		 * Create/refresh the list of available custom scene clips.
		 */
        public void RefreshClipLibrary()
        {
            if ((UnityEngine.Object) clipLibrary == (UnityEngine.Object) null) return;
            clipLibrary.RefreshClipIndex();
        }

        private void OnAtomRename(string oldID, string newID)
        {
            if (newID != containingAtom.uid) return;
            Log($"Containing Atom has been renamed to {newID}");
            if ((UnityEngine.Object) collections == (UnityEngine.Object) null) return;
            OnContainingAtomNameChanged.Invoke(new ContainingAtomNameChangedEventArgs
            {
                Before = oldID,
                After = newID
            });
        }
        #endregion

        #region Load / Save
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if (collections == null) return json;
            collections.Serialize(json, Storables.ClipCollections);
            needsStore = true;

            return json;
        }
        // private bool _skipBaseRestore;
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            // Test 21/10/2022
            // if (_skipBaseRestore == false)
            // {
                base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            // }
            // else
            // {
            //     Log(" - Late Restore");
            //     _skipBaseRestore = false;
            // }
            try
            {
                if (jc == null)
                {
                    Log("RestoreFromJSON: jc is null. Aborting restore!");
                    return;
                }
                if (jc.AsObject.HasKey(Storables.ClipCollections) == false)
                {
                    Log(jc);
                    Log("RestoreFromJSON:jc.AsObject.HasKey(Storables.ClipCollections) is false. Aborting restore!");
                    return;
                }
                var activeCollection = jc.AsObject.AsObject.HasKey(Storables.ActiveCollection)
                    ? jc.AsObject[Storables.ActiveCollection].ToString()
                    : "";
                var restoredJSON = jc.AsObject[Storables.ClipCollections];
                if (restoredJSON == null)
                {
                    Log("RestoreFromJSON: restoredJSON is null. Aborting restore!");
                    return;
                }
                SuperController.singleton.StartCoroutine(LoadDeferred(restoredJSON, activeCollection));
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(RestoreFromJSON)}: {e}");
            }
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true,
            bool setMissingToDefault = true)
        {
            Log($"LateRestoreFromJSON - ActiveCollection: {jc.AsObject[Storables.ActiveCollection]}");
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            // _skipBaseRestore = true;
            RestoreFromJSON(jc, restorePhysical, restoreAppearance, null, setMissingToDefault);
            // _skipBaseRestore = false;
        }

        public void Load(JSONNode json, string activeCollection = "")
        {
            if (collections == null)
            {
                insideRestore = false;
                return;
            }
            try
            {
                if (json == null || json.AsObject == null || $"{json}" == "{}")
                {
                    Log("Load called with empty json object: {json}. Aborting load!");
                    insideRestore = false;
                    return;
                }
                Log("Collections Parsing triggered.");

                collections.Parse(json, false, false, activeCollection);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(Load)}: {e}");
            }
            finally
            {
                insideRestore = false;
            }
        }

        private IEnumerator LoadDeferred(JSONNode json, string activeCollection = "")
        {
            if (insideRestore) yield break;
            insideRestore = true;
            Load(json, activeCollection);
        }
        #endregion

        #region Helpers
        public virtual IEnumerator CreateAtom(string atomType, string atomId, Action<Atom> onAtomCreated)
        {
            if (_atomCreationId != null && _atomCreationId == atomId) yield break;
            _atomCreationId = atomId;
            var atom = SuperController.singleton.GetAtomByUid(atomId);
            if (atom == null)
            {
                yield return SuperController.singleton.AddAtomByType(atomType, atomId);
                atom = SuperController.singleton.GetAtomByUid(atomId);
            }
            if (atom != null)
            {
                onAtomCreated(atom);
            }
            yield return null;
            _atomCreationId = null;
        }

        public void ParentLinkAtPosition(Atom atom, Atom parent, JSONStorable toPosition, string toPositionName)
        {
            atom.parentAtom = parent;
            var p = toPosition.transform.position;
            atom.mainController.currentPositionState = FreeControllerV3.PositionState.On;
            atom.mainController.transform.position = new Vector3(p.x, p.y, p.z);
            atom.mainController.SetLinkToAtom(parent.uid);
            atom.mainController.SetLinkToRigidbodyObject(toPositionName);
            atom.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            atom.mainController.canGrabPosition = false;
            atom.mainController.canGrabRotation = false;
        }

        #endregion

        #region Callbacks

        public void OnAddTriggerButtonClicked(string triggerType = TriggerManager.StartTriggerActions)
        {
            if (collections == null) return;
            if (collections.ActiveCollection == null) return;
            var triggerName = GetPlayClipTriggerName(TriggerColliderChooserJSON.val, collections.ActiveCollection.Name);
            if (string.IsNullOrEmpty(triggerName)) return;
            var receiverTargetName = CollectionManager.GetPlayClipActionName(collections.ActiveCollection.Name);
            Triggers.AddTriggerAction(TriggerColliderChooserJSON.val, triggerName, storeId, receiverTargetName, triggerType);
        }

        private void UpdateTriggerActionNamesOnCollectionRename(string oldCollectionName, string newCollectionName)
        {
            Log($"Renaming TriggerActions of Collection {oldCollectionName} to match Collection {newCollectionName}");
            if (string.IsNullOrEmpty(Tools.TrimAll(oldCollectionName)) || string.IsNullOrEmpty(Tools.TrimAll(newCollectionName))) return;
            foreach (var triggerCollider in TriggerColliderChooserJSON.choices)
            {
                // Rename PlayClip actions
                var oldTriggerActionName1 = GetPlayClipTriggerName(triggerCollider, oldCollectionName);
                var newTriggerActionName1 = GetPlayClipTriggerName(triggerCollider, newCollectionName);
                var newTriggerActionReceiverName1 = CollectionManager.GetPlayClipActionName(newCollectionName);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName1, newTriggerActionName1, newTriggerActionReceiverName1);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName1, newTriggerActionName1, newTriggerActionReceiverName1, TriggerManager.EndTriggerActions);

                // Rename PlayRandomClip actions
                var oldTriggerActionName2 = GetPlayRandomClipTriggerName(triggerCollider, oldCollectionName);
                var newTriggerActionName2 = GetPlayRandomClipTriggerName(triggerCollider, newCollectionName);
                var newTriggerActionReceiverName2 = CollectionManager.GetPlayRandomClipActionName(newCollectionName);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName2, newTriggerActionName2, newTriggerActionReceiverName2);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName2, newTriggerActionName2, newTriggerActionReceiverName2, TriggerManager.EndTriggerActions);

                // Rename QueueClip actions
                var oldTriggerActionName3 = GetQueueClipTriggerName(triggerCollider, oldCollectionName);
                var newTriggerActionName3 = GetQueueClipTriggerName(triggerCollider, newCollectionName);
                var newTriggerActionReceiverName3 = CollectionManager.GetQueueClipActionName(newCollectionName);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName3, newTriggerActionName3, newTriggerActionReceiverName3);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName3, newTriggerActionName3, newTriggerActionReceiverName3, TriggerManager.EndTriggerActions);

                // Rename QueueRandomClip actions
                var oldTriggerActionName4 = GetQueueRandomClipTriggerName(triggerCollider, oldCollectionName);
                var newTriggerActionName4 = GetQueueRandomClipTriggerName(triggerCollider, newCollectionName);
                var newTriggerActionReceiverName4 = CollectionManager.GetQueueRandomClipActionName(newCollectionName);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName4, newTriggerActionName4, newTriggerActionReceiverName4);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName4, newTriggerActionName4, newTriggerActionReceiverName4, TriggerManager.EndTriggerActions);
            }
        }

        private static string GetPlayClipTriggerName(string triggerCollider, string collectionName)
        {
            return Tools.TrimAll($"{triggerCollider}:{CollectionManager.GetPlayClipActionName(collectionName)}");
        }

        private  string GetQueueClipTriggerName(string triggerCollider, string collectionName)
        {
            return Tools.TrimAll($"{triggerCollider}:{CollectionManager.GetQueueClipActionName(collectionName)}");
        }

        private static string GetPlayRandomClipTriggerName(string triggerCollider, string collectionName)
        {
            return Tools.TrimAll($"{triggerCollider}:{CollectionManager.GetPlayRandomClipActionName(collectionName)}");
        }

        private  string GetQueueRandomClipTriggerName(string triggerCollider, string collectionName)
        {
            return Tools.TrimAll($"{triggerCollider}:{CollectionManager.GetQueueRandomClipActionName(collectionName)}");
        }

        public void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs eventArgs)
        {
            if (collections == null) return;

            var atom = collections.ActiveCollection?.ReceiverAtom;
            if (atom == null) return;
            ReceivingAtomJSON.val = atom;

            var node = collections.ActiveCollection?.ReceiverNode;
            if (node == null) return;
            ReceivingNodeJSON.val = node;
        }

        public void OnActiveCollectionNameChanged(CollectionManager.ActiveCollectionNameChangedEventArgs eventArgs)
        {
            UpdateTriggerActionNamesOnCollectionRename(eventArgs.Before, eventArgs.After);
        }

        /**
		* Load audio source ReceiverAtom.
		*/
        private void SetReceivingAtom(string uid)
        {
            if (Tools.IsEmptyChoice(uid) || (UnityEngine.Object) collections == (UnityEngine.Object) null || collections.ActiveCollection == null) return;
            try
            {
                receivingAtom = SuperController.singleton.GetAtomByUid(uid);
                if (receivingAtom == null)
                {
                    Log($"Audio Source {uid} does not exist.");
                    return;
                }
                var atomChanged = collections.ActiveCollection.ReceiverAtom != receivingAtom.uid;
                collections.ActiveCollection.ReceiverAtom = receivingAtom.uid;
                if (!atomChanged) return;
                var guessedReceiver = GuessInitialReceivingNode(receivingAtom);
                insideRestore = true;
                ReceivingNodeJSON.val = guessedReceiver;
                insideRestore = false;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(SetReceivingAtom)}: {e}");
            }
        }

        /**
		* Load audio source receiver.
		*/
        private void SetReceivingNode(string uid)
        {
            if (Tools.IsEmptyChoice(uid) || (UnityEngine.Object) collections == (UnityEngine.Object) null || collections.ActiveCollection == null) return;
            receivingNode = SuperController.singleton.GetAtomByUid(ReceivingAtomJSON.val)?.GetStorableByID(uid);
            if (receivingNode == null)
            {
                _missingReceiverNodeStoreId = uid;
                return;
            }

            insideRestore = true;
            ReceivingNodeJSON.val = receivingNode.name;
            insideRestore = false;
            collections.ActiveCollection.ReceiverNode = receivingNode.name;
            collections.ActiveCollection.SyncAudioReceiver();
        }

        private void FindMissingReceiverNode() {
            if (_missingReceiverNodeStoreId != "" && receivingAtom != null) {
                var missingReceiver = receivingAtom.GetStorableByID(_missingReceiverNodeStoreId);
                if (missingReceiver == null) return;
                SetReceivingNode(_missingReceiverNodeStoreId);
                _missingReceiverNodeStoreId = "";
            }
        }
        #endregion

        #region Debug
        /**
         * Custom debug logging
         */
        public void Log(string message)
        {
            #if DEBUG
            SuperController.LogMessage($"[{this.storeId}]: {message}");
            #endif
        }

        /* public void LogObject(object obj)
        {
            #if DEBUG
            SuperController.LogMessage(JsonConvert.SerializeObject(obj));
            #endif
        } */

        /**
         * UI debugging. Depends on VAM Dev Tools by Acidbubbles (https://github.com/acidbubbles/vam-devtools).
         */
        public void DebugObject(Transform obj)
        {
            #if DEBUG
            DebugObject(obj.gameObject);
            #endif
        }

        public void DebugObject(GameObject obj)
        {
            #if DEBUG
            containingAtom.BroadcastMessage("DevToolsInspectUI", obj);
            #endif
        }
        #endregion
    }
}
