﻿using System;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class ClipUI
    {
        public bool IsInitialised;
        public Transform Container;
        public UIDynamicButton PreviewButton;
        public UIDynamicButton ToggleButton;
        public Outline ToggleOutline;

        public bool SetInCollectionState(bool state)
        {
            if (ToggleButton == null) return state;
            ToggleButton.textColor = state ? Styles.EnabledText : Styles.DisabledText;
            ToggleButton.buttonColor = state ? Styles.EnabledBg : Styles.DisabledBg;
            if (ToggleOutline != null) ToggleOutline.effectColor = state ? Color.white : Styles.DisabledBg;

            return state;
        }

        public bool SetCursorState(bool state)
        {
            if (PreviewButton == null) return state;
            PreviewButton.textColor = state ? Styles.CursorEnabledText : Styles.CursorDisabledText;
            PreviewButton.buttonColor = state ? Styles.CursorEnabledBg : Styles.CursorDisabledBg;
            return state;
        }

        public void Destroy()
        {
            try
            {
                if (Container != null) Object.Destroy(Container.gameObject);
                Container = null;
                if (PreviewButton != null) Object.Destroy(PreviewButton.gameObject);
                PreviewButton = null;
                if (ToggleButton != null) Object.Destroy(ToggleButton.gameObject);
                ToggleButton = null;
                if (ToggleOutline != null) Object.Destroy(ToggleOutline.gameObject);
                ToggleOutline = null;
                IsInitialised = false;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(ClipUI)}.{nameof(Destroy)}: {e}");
            }
        }
    }

    public class AudioMateClip
    {
        public string DisplayName => AudioClip?.displayName;

        public string UID
        {
            get {
                if (_uid != null || AudioClip == null) return _uid;
                _uid = $"AudioMateClip_{Tools.TrimFileExtension(AudioClip.uid)}";

                return _uid;
            }
        }

        private string _uid;
        public NamedAudioClip AudioClip { get; private set; }
        public AudioSourceControl Receiver;

        public bool IsInActiveCollection { get; private set; }
        public bool HasCursor { get; private set; }

        public ClipUI UI;

        public bool IsProvisional;

        public string SourcePath;

        public AudioMateClip(NamedAudioClip audioClip)
        {
            AudioClip = audioClip;
            Init();
        }

        public AudioMateClip(JSONNode jn)
        {
            FromJSON(jn);
            Init();
        }

        private void Init()
        {
            UI = new ClipUI();
            IsInActiveCollection = false;
            HasCursor = false;
            AddSourcePath();
        }

        public void RefreshUI()
        {
            if (UI == null || UI.IsInitialised == false) return;
            UI.SetInCollectionState(IsInActiveCollection);
            UI.SetCursorState(HasCursor);
        }

        public JSONClass ToJSON()
        {
            return new JSONClass
            {
                { "uid", UID },
                { "audioClip", AudioClip?.uid },
                { "hasCursor", HasCursor.ToString() },
                { "sourcePath", SourcePath }
            };
        }

        private void FromJSON(JSONNode jn)
        {
            if (jn == null || !jn.Childs.Any()) return;
            var sourceClip = jn["audioClip"] ?? jn["sourceClip"] ?? jn["sourceClipUID"]; // Check for legacy formats
            SourcePath = jn["sourcePath"];
            AudioClip = sourceClip != null ? URLAudioClipManager.singleton.GetClip(sourceClip) : null;
            if (AudioClip == null && SourcePath != null)
            {
                URLAudioClipManager.singleton.QueueClip(SourcePath);
            }
            if (string.IsNullOrEmpty(SourcePath))
            {
                AddSourcePath();
            }
        }

        private void AddSourcePath()
        {
            if (AudioClip == null) return;
            var sceneClipData = URLAudioClipManager.singleton.GetJSON()["clips"];
            foreach (var clipData in sceneClipData.Childs)
            {
                if (clipData["displayName"].Value == AudioClip.displayName)
                {
                    SourcePath = clipData["url"];
                }
            }
        }

        /**
         * Play the assigned audio clip with optional "if clear" param.
         */
        public bool Play(AudioSourceControl receiver = null, bool ifClear = false, bool clearQueue = false)
        {
            if (AudioClip == null) return false;
            if ((UnityEngine.Object) receiver != (UnityEngine.Object) null) Receiver = receiver;
            if (ifClear && Receiver.audioSource.isPlaying) return false;
            if (clearQueue)
            {
                Receiver.PlayNowClearQueue(AudioClip);
                return true;
            }

            Receiver.PlayNow(AudioClip);
            return true;
        }

        /**
         * Add assigned audio clip to audio _c queue list.
         */
        public void Queue(AudioSourceControl receiver = null)
        {
            if (AudioClip == null) return;
            if ((UnityEngine.Object) receiver != (UnityEngine.Object) null)
            {
                Receiver = receiver;
            }
            else
            {
                return;
            }
            Receiver.QueueClip(AudioClip);
        }

        public void DeleteFromScene()
        {
            if (AudioClip == null) return;
            // TODO
        }

        public void Destroy()
        {
            try
            {
                UI.Destroy();
                Receiver = null;
                AudioClip = null;
            } catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClip)}.{nameof(Destroy)}: {e}");
            }
        }

        public void DestroyUI()
        {
            try
            {
                UI.Destroy();
            } catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClip)}.{nameof(DestroyUI)}: {e}");
            }
        }

        public bool ToggleState()
        {
            IsInActiveCollection = !IsInActiveCollection;
            UI.SetInCollectionState(IsInActiveCollection);
            RefreshUI();
            return IsInActiveCollection;
        }

        public bool SetState(bool state)
        {
            IsInActiveCollection = state;
            UI.SetInCollectionState(state);
            RefreshUI();
            return state;
        }

        public bool SetCursor(bool state)
        {
            HasCursor = state;
            UI.SetCursorState(state);
            RefreshUI();
            return state;
        }
    }
}
