﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{
    public class ClipCursor
    {
        public int Index;
        private AudioMateClip _clip;
        private readonly AudioMateController _c;

        public ClipCursor(AudioMateController controller)
        {
            _c = controller;
        }

        // Set Clip Cursor
        public void Set(AudioMateClip newCursorClip)
        {
            if (_c == null) return;
            Reset();
            _clip = newCursorClip;
            Index = _c.Clips.IndexOf(_clip);

            _clip.SetCursor(true);
        }

        private void Reset()
        {
            _c.Clips.ForEach(clip =>
            {
                clip.SetCursor(false);
            });
        }
    }

    public class ClipLibrary : MonoBehaviour
    {
        public class ClipsRefreshedEvent : UnityEvent { }
        public readonly ClipsRefreshedEvent OnClipsRefreshed = new ClipsRefreshedEvent();

        private AudioMateController _c;

        private List<NamedAudioClip> _sourceClipList;
        private bool _processing;

        public List<AudioMateClip> Clips { get; private set; }

        public ClipCursor Cursor;
        public bool isInitialised;
        private bool _isBound;

        public void Init(AudioMateController controller)
        {
            if (controller == null) return;
            _c = controller;
            Cursor = new ClipCursor(_c);
            Clips = new List<AudioMateClip>();
            _c.fileManager.OnSceneFilesUpdated.AddListener(OnSourceClipsUpdated);
        }

        public void Bind()
        {
            Log("Binding Clip Library");
            _c.collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);
            _c.collections.OnActiveCollectionUpdated.AddListener(OnActiveCollectionUpdated);
            _c.fileManager.OnSceneFilesUpdated.AddListener(OnSourceClipsUpdated);
            _isBound = true;
            _processing = false;
            RefreshClipIndex();
        }

        public void SetCursorIndex(int index)
        {
            Cursor?.Set(Clips[index]);
        }

        public void RefreshClipIndex()
        {
            if (_isBound == false) return;
            Log(" - RefreshClipIndex");
            try
            {
                SuperController.singleton.StartCoroutine(DeferredIndexing());
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(ClipLibrary)}.{nameof(RefreshClipIndex)}: {e}");
            }
        }

        private IEnumerator DeferredIndexing()
        {
            Log($"DeferredIndexing");
            if (_processing)
            {
                Log($"DeferredIndexing - Already processing. Aborting.");
                yield break;
            }
            _processing = true;

            /* if (this == null)
            {
                _processing = false;
                Log($"DeferredIndexing - this is null. Aborting.");
                yield break;
            } */
            _sourceClipList = URLAudioClipManager.singleton.GetCategoryClips("web");
            // Convert new clips to AudioMateClips and add them to the list
            var clipCount = 0;
            if (_sourceClipList != null)
            {
                foreach (var sourceClip in _sourceClipList)
                {
                    var audioMateClip = Clips.Find(x => x.AudioClip == sourceClip);
                    if (audioMateClip != null) continue;
                    audioMateClip = new AudioMateClip(sourceClip);
                    Clips.Add(audioMateClip);
                    clipCount++;
                }

                foreach (var audioMateClip in Clips)
                {
                    if (audioMateClip.AudioClip == null)
                    {
                        audioMateClip.Destroy();
                    }
                    var sourceClipExists = _sourceClipList.Exists(sourceClip => sourceClip == audioMateClip.AudioClip);
                    if (!sourceClipExists) audioMateClip.Destroy();
                }
            }
            else
            {
                Log($"DeferredIndexing - _sourceClipList is null.");
            }
            Log($"Added {clipCount} Clips to library.");

            isInitialised = true;
            OnClipsRefreshed.Invoke();
            _processing = false;
        }

        public void RemoveAllClips()
        {
            URLAudioClipManager.singleton.RemoveAllClips();
            _c.fileManager.OnSceneFilesUpdated.Invoke();
        }

        public void RemoveUnusedClips()
        {
            var unusedClips = _c.collections.UnusedClips();
            unusedClips.ForEach(clip =>
            {
                URLAudioClipManager.singleton.RemoveClip(clip.AudioClip);
            });
            _c.fileManager.OnSceneFilesUpdated.Invoke();
        }

        public AudioMateClip FindByAudioClipUID(string audioClipUID)
        {
            return Clips.Find(x => x.AudioClip.uid == audioClipUID);
        }

        public void ProcessProvisionalClips()
        {
            Log($"Processing provisional Clips of {_c.collections.CollectionsWithProvisionalClips.Count().ToString()} Collection(s).");

            foreach (var collection in _c.collections.CollectionsWithProvisionalClips)
            {
                Log($"Processing collection {collection.Name}.");
                var isActive = collection == _c.collections.ActiveCollection;
                if (!collection.GetProvisionalClips().Any())
                {
                    Log($"Skipping {collection.Name} because it doesn't have any provisional Clips anymore.");
                    continue;
                }
                foreach (var provisionalClip in collection.GetProvisionalClips())
                {
                    if (provisionalClip.AudioClip == null)
                    {
                        continue;
                    }
                    var audioMateClip = FindByAudioClipUID(provisionalClip.AudioClip.uid);
                    if (audioMateClip == null)
                    {
                        Log($"Couldn't find source file {provisionalClip.AudioClip.uid}. Skipping.");
                        continue;
                    }
                    if (isActive) audioMateClip.SetState(true);
                    collection.SwapProvisionalClip(provisionalClip, audioMateClip);
                }
            }
            OnClipsRefreshed.Invoke();
        }

        public void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
            // ...
        }

        public void OnActiveCollectionUpdated()
        {
            // ...
        }

        public void OnSourceClipsUpdated()
        {
            Log("Scene Clips updated!");
            RefreshClipIndex();
        }

        public void OnDestroy()
        {
            _isBound = false;
            try
            {
                _processing = false;
            }
            catch (Exception e)
            {
                SuperController.LogMessage($"AudioMate: [UIClipLibrary.OnDestroy] {e}");
            }
        }

        private void Log(string message)
        {
            if (_c == null) return;
            _c.Log($" [ClipLibrary]: {message}");
        }
    }
}
