﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{
    public class CollectionManager : MonoBehaviour
    {

        private class CollectionActions
        {
            public JSONStorableAction PlayAction;           // Plays the collection according to the user's collection settings
            public JSONStorableAction QueueAction;          // Queues the collection according to the user's collection settings
            public JSONStorableAction PlayRandomAction;     // Always plays a random clip from the collection, ignores shuffle being set to false
            public JSONStorableAction QueueRandomAction;    // Always plays a random clip from the collection, ignores shuffle being set to false
        }

        public class ActiveCollectionSelectedEventArgs { public Collection Before; public Collection After; }
        public class ActiveCollectionNameChangedEventArgs { public string Before; public string After; }

        public class ActiveCollectionSelectedEvent : UnityEvent<ActiveCollectionSelectedEventArgs> { }
        public class ActiveCollectionUpdatedEvent : UnityEvent { }
        public class ActiveCollectionNameChangedEvent : UnityEvent<ActiveCollectionNameChangedEventArgs> { }

        public readonly ActiveCollectionSelectedEvent OnActiveCollectionSelected = new ActiveCollectionSelectedEvent();
        public readonly ActiveCollectionUpdatedEvent OnActiveCollectionUpdated = new ActiveCollectionUpdatedEvent();
        public readonly ActiveCollectionNameChangedEvent OnActiveCollectionNameChanged = new ActiveCollectionNameChangedEvent();

        private AudioMateController _c;
        private readonly List<Collection> _collections = new List<Collection>();

        public IEnumerable<Collection> CollectionsWithProvisionalClips =>
            _collections.FindAll(x => x.GetProvisionalClips().Count > 0);

        private readonly Dictionary<string,CollectionActions> _collectionActions = new Dictionary<string, CollectionActions>();

        public JSONStorableStringChooser CollectionsJSON;

        public Collection ActiveCollection { get; private set; }

        public int Count => _collections.Distinct().Count();

        public IEnumerable<string> Names => _collections.Select(collection => collection.Name).Distinct().ToList();
        public bool isInitialized;
        public bool isParsingCollections;
        public bool isAutoPlayRoutineRunning;
        private Coroutine _autoPlayRoutine;

        #region Initialization
        public void Init(AudioMateController controller)
        {
            if ((UnityEngine.Object) controller == (UnityEngine.Object) null)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(Init)}: Controller is null.");
                return;
            }
            _c = controller;
            if (isInitialized)
            {
                Log("Collection manager is already initialized");
                return;
            }
            Log("Initializing Collection Manager");
            if ((UnityEngine.Object) _c.ui == (UnityEngine.Object) null)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(Init)}: UI is not initialized.");
                return;
            }
            _c.OnContainingAtomNameChanged.AddListener(OnAtomRename);
            CollectionsJSON = _c.CollectionsJSON;
            CollectionsJSON.setCallbackFunction = SelectActiveCollection;
            SelectActiveCollection();
            isInitialized = true;
        }

        public void StartAutoPlayRoutine()
        {
            if (isAutoPlayRoutineRunning || _collections == null) return;
            if (_autoPlayRoutine != null) SuperController.singleton.StopCoroutine(_autoPlayRoutine);
            _autoPlayRoutine = SuperController.singleton.StartCoroutine(AutoPlayDeferred());
        }

        private IEnumerator AutoPlayDeferred()
        {
            while (true)
            {
                if (this == null) yield break;
                if (isAutoPlayRoutineRunning == false) isAutoPlayRoutineRunning = true;
                foreach (var audioMateClipCollection in _collections.FindAll(x => x.AutoPlay && !x.HasAutoPlayBeenTriggered))
                {
                    audioMateClipCollection.HasAutoPlayBeenTriggered = true;
                    audioMateClipCollection.PlayClip();
                }
                foreach (var audioMateClipCollection in _collections.FindAll(x => x.ContinuousPlay))
                {
                    if (!audioMateClipCollection.IsPlaying() && audioMateClipCollection.HasPlayedAnything) audioMateClipCollection.PlayClip();
                }

                yield return new WaitForSeconds(.2f);
            }
        }

        private void AddDefaultCollection()
        {
            if (_collections == null) return;

            var newCollection = AddCollection();
            SelectActiveCollection(newCollection);
        }
        #endregion

        #region Helper
        private string GetAvailableDefaultName()
        {
            var newName = "Untitled";
            var i = 2;
            while (Names.Contains(newName))
            {
                newName = "Untitled" + i;
                i++;
            }

            return newName;
        }

        public void SyncCollectionNames()
        {
            var choices = new List<string>();
            choices.AddRange(Names);
            CollectionsJSON.choices = choices;
        }
        #endregion

        #region Collection Management

        public Collection AddCollection(string collectionName = null)
        {
            if (collectionName == null)
            {
                collectionName = GetAvailableDefaultName();
            }
            var newCollection = new Collection(collectionName);
            return AddCollection(newCollection);
        }

        private Collection AddCollection(Collection collection)
        {
            if (collection == null || string.IsNullOrEmpty(collection.Name)) return null;
            if (collection.ReceiverAtom == null) collection.ReceiverAtom = _c.containingAtom.uid;
            if (collection.ReceiverNode == null) collection.ReceiverNode = AudioMateController.GuessInitialReceivingNode(_c.containingAtom);
            collection.OnClipPlayed.AddListener(() => OnClipPlayed(collection));
            collection.VolumeControlEnabled = _c.VolumeControlEnabledJSON.val;
            collection.PitchControlEnabled = _c.PitchControlEnabledJSON.val;
            var existingName = _collections.FindLastIndex(x => x.Name == collection.Name);
            if (existingName == -1)
            {
                Log($"Adding Collection {collection.Name}");
                _collections.Add(collection);
            }
            else
            {
                Log($"Inserting Collection {collection.Name}");
                _collections.Insert(existingName, collection);
            }
            AddNewCollectionActions(collection.Name);
            if (!CollectionsJSON.choices.Contains(collection.Name))
            {
                Log($"Adding Name {collection.Name}");
                CollectionsJSON.choices.Add(collection.Name);
            }
            SelectActiveCollection(collection);

            return collection;
        }

        private void OnClipPlayed(Collection collection)
        {
            var index = _c.Clips.IndexOf(collection.LastPlayedClip);
            _c.clipLibrary.SetCursorIndex(index);
        }

        public bool RemoveCollection(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName)) return false;
            var collectionToRemove = _collections.Find(x => x.Name == collectionName);
            return collectionToRemove != null && RemoveCollection(collectionToRemove);
        }

        private bool RemoveCollection(Collection collection)
        {
            var collectionName = collection?.Name;
            if (string.IsNullOrEmpty(collectionName)) return false;
            CollectionsJSON.choices.Remove(collectionName);
            RemoveCollectionActions(collectionName);
            var result = _collections.Remove(collection);
            SyncCollectionNames();
            return result;
        }

        public bool RemoveActiveCollection()
        {
            var collectionIndex = _collections.IndexOf(ActiveCollection);
            var result = RemoveCollection(ActiveCollection);
            if (_collections.Count == 0)
            {
                AddDefaultCollection();
            }
            else
            {
                collectionIndex--;
                SelectActiveCollection(collectionIndex == -1 ? _collections.Last() : _collections[collectionIndex]);
            }
            return result;
        }

        public void SelectActiveCollection(Collection collection = null)
        {
            Log($"Selecting new active collection{(collection != null ? $" {collection.Name}" : " (No name has been set)")}");
            while (true)
            {
                if (collection != null)
                {
                    var previous = ActiveCollection;
                    ActiveCollection = collection;
                    ActiveCollection.SyncAudioReceiver();
                    if (CollectionsJSON.val != ActiveCollection.Name) CollectionsJSON.val = ActiveCollection.Name;
                    OnActiveCollectionSelected.Invoke(new ActiveCollectionSelectedEventArgs {Before = previous, After = ActiveCollection});
                    break;
                }

                if (_collections.Count > 0)
                {
                    collection = _collections.First();
                    continue;
                }

                AddDefaultCollection();
                collection = null;
            }
        }

        public void SelectActiveCollection(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName))
            {
                Log($"Can't select active Collection. Collection name is missing. Aborting.");
                return;
            }
            var collection = _collections.Find(x => x.Name == collectionName);
            if (collection == null)
            {
                return;
            }
            Log($"Selecting {collectionName} as active Collection");
            SelectActiveCollection(collection);
        }

        public void Enable(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.Enable();
        }

        public void Enable()
        {
            ActiveCollection?.Enable();
        }

        public void Disable(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.Disable();
        }

        public void Disable()
        {
            ActiveCollection?.Disable();
        }

        public void EnableAll()
        {
            foreach (var collection in _collections)
            {
                collection.Enable();
            }
        }

        public void DisableAll()
        {
            foreach (var collection in _collections)
            {
                collection.Disable();
            }
        }

        public void SetActiveCollectionName(string newCollectionName)
        {
            if (string.IsNullOrEmpty(newCollectionName)) return;
            if (ActiveCollection == null) return;
            var oldCollectionName = ActiveCollection.Name;
            ActiveCollection.Name = newCollectionName;
            SyncCollectionNames();
            CollectionsJSON.val = newCollectionName;
            UpdateCollectionActionNames(oldCollectionName, newCollectionName);

            OnActiveCollectionNameChanged.Invoke(new ActiveCollectionNameChangedEventArgs
            {
                Before = oldCollectionName,
                After = newCollectionName
            });
        }

        /**
         * Adds a clip to the active collections
         */
        public void AddClipToActiveCollection(AudioMateClip clip, bool noEvent = false)
        {
            ActiveCollection?.Add(clip);
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds a range of clips to the active collections
         */
        public void AddClipRangeToActiveCollection(int fromIndex, int toIndex, bool noEvent = false)
        {
            if (ActiveCollection == null) return;
            for (var i = fromIndex; i <= toIndex; i++)
            {
                var clip = _c.clipLibrary.Clips.ElementAtOrDefault(i);
                if (clip == null) continue;
                ActiveCollection.Add(clip);
                _c.clipLibrary.SetCursorIndex(i);
            }
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Removes a clip from a collection
         */
        public bool RemoveClipFromActiveCollection(AudioMateClip clip, bool noEvent = false)
        {
            if (ActiveCollection == null || !ActiveCollection.Remove(clip)) return false;
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
            return true;

        }

        /**
         * Adds all available clips to the active collections
         */
        public void AddAllClipsToActiveCollection()
        {
            foreach (var clip in _c.clipLibrary.Clips)
            {
                AddClipToActiveCollection(clip, true);
            }

            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds 10 clip to the active collections, starting from the cursor
         */
        public void Add10ClipsToActiveCollection()
        {
            AddClipRangeToActiveCollection(_c.clipLibrary.Cursor.Index, _c.clipLibrary.Cursor.Index + 10, true);
            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds 20 clip to the active collections, starting from the cursor
         */
        public void Add20ClipsToActiveCollection()
        {
            AddClipRangeToActiveCollection(_c.clipLibrary.Cursor.Index, _c.clipLibrary.Cursor.Index + 20, true);
            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Removes all clips from the active collections
         */
        public void ClearActiveCollection()
        {
            ActiveCollection?.Clear();
            OnActiveCollectionUpdated.Invoke();
        }

        public void ClearAllCollections()
        {
            _collections.ForEach(collection => collection.Clear());
            OnActiveCollectionUpdated.Invoke();
        }

        public void SetVolumeControl(bool val)
        {
            _collections.ForEach(collection => collection.VolumeControlEnabled = val);
        }

        public void SetPitchControl(bool val)
        {
            _collections.ForEach(collection => collection.PitchControlEnabled = val);
        }
        #endregion

        #region Clip Actions

        /**
         * Plays a clip from a specific Collection. If shuffle mode is enabled, a random clip will be played.
         */
        public void PlayClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            if (queue)
            {
                collection?.QueueClip();
            }
            else
            {
                collection?.PlayClip();
            }
        }

        /**
         * Queues a clip from a specific Collection. If shuffle mode is enabled, a random clip will be queued.
         */
        public void QueueClipFromCollection(string collectionName)
        {
            PlayClipFromCollection(collectionName, true);
        }

        /**
         * Plays a random clip from a specific Collection
         */
        public void PlayRandomClipFromCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            if (queue)
            {
                collection?.QueueRandomClip();
            }
            else
            {
                collection?.PlayRandomClip();
            }
        }

        /**
         * Queues a random clip from a specific Collection
         */
        public void QueueRandomClipFromCollection(string collectionName)
        {
            PlayRandomClipFromCollection(collectionName, true);
        }

        public void PlayClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayClip();
            }
            else
            {
                PlayClipFromCollection(collectionName);
            }
        }

        public void QueueClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.QueueClip();
            }
            else
            {
                QueueClipFromCollection(collectionName);
            }
        }

        public void PlayRandomClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayRandomClip();
            }
            else
            {
                PlayRandomClipFromCollection(collectionName);
            }
        }

        public void QueueRandomClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.QueueRandomClip();
            }
            else
            {
                QueueRandomClipFromCollection(collectionName);
            }
        }

        public IEnumerable<AudioMateClip> UsedClips(string collectionFilter = null)
        {
            var result = new List<AudioMateClip>();

            if (collectionFilter == null)
            {
                foreach (var collection in _collections)
                {
                    result.AddRange(collection.Clips);
                }
            }
            else
            {
                var collection = _collections.Find(x => x.Name == collectionFilter);
                if (collection == null) return result;
                result = collection.Clips;
            }

            return result;
        }

        public List<AudioMateClip> UnusedClips(string collectionFilter = null)
        {
            var result = new List<AudioMateClip>();
            result.AddRange(_c.Clips);
            if (collectionFilter == null)
            {
                foreach (var clip in _collections.SelectMany(collection => collection.Cast<AudioMateClip>()))
                {
                    result.Remove(clip);
                }
            }
            else
            {
                var collection = _collections.Find(x => x.Name == collectionFilter);
                if (collection == null) return result;
                foreach (AudioMateClip clip in collection)
                {
                    result.Remove(clip);
                }
            }
            return result;
        }
        #endregion

        /**
         * Creates all needed Actions for a new collection
         */
        public void AddNewCollectionActions(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName))
            {
                return;
            }
            if (_collectionActions == null) return;
            if (_collectionActions.ContainsKey(collectionName)) return;
            var newActions = new CollectionActions
            {
                PlayAction = new JSONStorableAction(GetPlayClipActionName(collectionName),
                    () => PlayClipAction(collectionName)),
                QueueAction = new JSONStorableAction(GetQueueClipActionName(collectionName),
                    () => QueueClipAction(collectionName)),
                PlayRandomAction = new JSONStorableAction(GetPlayRandomClipActionName(collectionName),
                    () => PlayClipAction(collectionName)),
                QueueRandomAction = new JSONStorableAction(GetQueueRandomClipActionName(collectionName),
                    () => QueueClipAction(collectionName))
            };

            _c.RegisterAction(newActions.PlayAction);
            _c.RegisterAction(newActions.QueueAction);
            _c.RegisterAction(newActions.PlayRandomAction);
            _c.RegisterAction(newActions.QueueRandomAction);
            _collectionActions.Add(collectionName, newActions);
        }

        private void UpdateCollectionActionNames(string oldCollectionName, string newCollectionName)
        {
            if (string.IsNullOrEmpty(oldCollectionName) || string.IsNullOrEmpty(newCollectionName))
            {
                Log($"{nameof(AddNewCollectionActions)}: one or both input values are empty");
                return;
            }
            if (_collectionActions == null) return;
            CollectionActions actions;
            if (!_collectionActions.TryGetValue(oldCollectionName, out actions)) return;
            if (actions != null)
            {
                RemoveCollectionActions(oldCollectionName);
            }
            AddNewCollectionActions(newCollectionName);
        }

        private void RemoveCollectionActions(string collectionName)
        {
            CollectionActions actions;
            if (!_collectionActions.TryGetValue(collectionName, out actions)) return;
            _c.DeregisterAction(actions.PlayAction);
            _c.DeregisterAction(actions.QueueAction);
            _collectionActions.Remove(collectionName);
            actions.PlayAction = null;
            actions.QueueAction = null;
        }

        public static string GetPlayClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayClipFromCollectionAction}{collectionName}");
        }

        public static string GetQueueClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.QueueClipFromCollectionAction}{collectionName}");
        }

        public static string GetPlayRandomClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayRandomClipFromCollectionAction}{collectionName}");
        }

        public static string GetQueueRandomClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.QueueRandomClipFromCollectionAction}{collectionName}");
        }

        #region Callbacks

        public void OnAtomRename(AudioMateController.ContainingAtomNameChangedEventArgs args)
        {
            var oldID = args.Before;
            var newID = args.After;
            foreach (var collection in _collections.Where(collection => collection.ReceiverAtom == oldID))
            {
                collection.ReceiverAtom = newID;
                collection.SyncAudioReceiver();
            }
        }

        public void OnDestroy()
        {
            if (_autoPlayRoutine != null) SuperController.singleton.StopCoroutine(_autoPlayRoutine);
            _autoPlayRoutine = null;
        }
        #endregion

        #region Serializer

        public void Serialize(JSONClass json, string id, string collectionName = null,  bool resetAudioSourceIds = false)
        {
            if (json == null) json = new JSONClass {{id, null}};
            if (json[id] == null) json.Add(id, null);
            json[id] = GetCollectionsJSON(collectionName, resetAudioSourceIds);
        }

        private JSONClass GetCollectionsJSON(string collectionName = null, bool resetAudioSourceIds = false)
        {
            var json = new JSONClass();
            var receiverAtomId = _c.containingAtom.uid;
            try
            {
                var query = _collections.Where(collection => collection != null && collection.Count != 0);
                if (collectionName != null) query = query.Where(collection => collection.Name == collectionName);
                foreach (var collection in query)
                {
                    json[collection.Name] = collection.ToJSON();
                    if (resetAudioSourceIds && json[collection.Name]["receiverAtom"].Value == receiverAtomId)
                    {
                        json[collection.Name]["receiverAtom"] = "Person";
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(GetCollectionsJSON)}: {e}");
            }
            return json;
        }

        public void Parse(JSONNode json, bool merge = false, bool createMissingAudioSources = false, string activeCollection = "")
        {
            Log($"Parse JSON, activeCollection is {activeCollection}");
            if (json == null || !json.Childs.Any() || isParsingCollections) return;
            try
            {
                _c.StartCoroutine(ParseCollections(json.Childs.ToList(), merge, createMissingAudioSources, activeCollection));
            }
            catch (Exception e)
            {
                SuperController.LogError(
                    $"AudioMate.{nameof(CollectionManager)}.{nameof(Parse)}: {e}");
            }
        }

        private IEnumerator ParseCollections(List<JSONNode> collectionList, bool merge, bool createMissingAudioSources, string activeCollection = "")
        {
            if (isParsingCollections) yield break;
            isParsingCollections = true;
            DisableAll();
            if (merge == false) _collections.Clear();
            foreach (var collectionJSON in collectionList)
            {
                Log($"Parsing collection {collectionJSON["name"]}");
                var mergeTarget = _collections.Find(x => x.Name == collectionJSON["name"]);
                var collection = mergeTarget ?? new Collection(collectionJSON["name"]);
                if (!collection.Parse(collectionJSON, merge, true) || mergeTarget != null) continue;
                if (collection.ReceiverAtom != null && createMissingAudioSources)
                {
                    var audioSource = SuperController.singleton.GetAtomByUid(collection.ReceiverAtom);
                    if (audioSource == null)
                    {
                        _c.StartCoroutine(_c.CreateAtom("AudioSource", collection.ReceiverAtom, newAtom =>
                        {
                            var hip = _c.containingAtom.GetStorableByID("hip");
                            _c.ParentLinkAtPosition(newAtom, _c.containingAtom, hip, "hip");
                            audioSource = newAtom;
                            var audioSourceControl = audioSource.GetStorableByID("AudioSource") as AudioSourceControl;
                            if (audioSourceControl == null) return;
                            audioSourceControl.spatialize = true;
                            audioSourceControl.volume = 1.0f;
                            audioSourceControl.pitch = 1.0f;
                        }));
                    }
                }

                yield return null;
                collection.Init();
                AddCollection(collection);
            }

            yield return null;
            EnableAll();
            SelectActiveCollection(activeCollection == "" ? _collections.First().Name : activeCollection);
            _c.needsProvisionalClipProcessing = true;
            isParsingCollections = false;
        }

        #endregion

        private void Log(string message)
        {
            if ((UnityEngine.Object) _c == (UnityEngine.Object) null) return;
            _c.Log($"CollectionManager: {message}");
        }

    }
}
