﻿using System;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace AudioMate
{
    public class TriggerManager
    {
        private readonly AudioMateController _c;
        private readonly List<string> _colliderChoices = new List<string>();

        public List<string> ColliderChoices
        {
            get {
                _colliderChoices.Clear();
                _colliderChoices.Add("LipTrigger");
                _colliderChoices.Add("MouthTrigger");
                _colliderChoices.Add("ThroatTrigger");
                _colliderChoices.Add("lNippleTrigger");
                _colliderChoices.Add("rNippleTrigger");
                _colliderChoices.Add("LabiaTrigger");
                _colliderChoices.Add("VaginaTrigger");
                _colliderChoices.Add("DeepVaginaTrigger");
                _colliderChoices.Add("DeeperVaginaTrigger");
                // Add custom collision triggers
                _colliderChoices.AddRange(from atom in SuperController.singleton.GetAtoms()
                    where atom != null
                    where atom.type == "CollisionTrigger"
                    select atom.name);
                return _colliderChoices;
            }
        }

        public const string DefaultCollider = "VaginaTrigger";
        public const string StartTriggerActions = "startActions";
        public const string EndTriggerActions = "endActions";

        public TriggerManager(AudioMateController controller)
        {
            _c = controller;
            Init();
        }

        private void Init()
        {
        }

        private static JSONStorable GetPluginStorableById(Atom atom, string id)
        {
            var storableIdName = atom.GetStorableIDs().FirstOrDefault((string storeId) => !string.IsNullOrEmpty(storeId) && storeId.Contains(id));
            return storableIdName == null ? null : atom.GetStorableByID(storableIdName);
        }

        private void CleanUp()
        {
            // ...
        }

        private static bool DoesTriggerActionExist(CollisionTrigger trigger, string triggerActionName, string triggerActionType = StartTriggerActions)
        {
            JSONNode presentTriggers = trigger.trigger.GetJSON();
            var asArray = triggerActionType == StartTriggerActions ? presentTriggers["startActions"].AsArray : presentTriggers["endActions"].AsArray;
            for (var i = 0; i < asArray.Count; i++) {
                var asObject = asArray[i].AsObject;
                string name = asObject["name"];
                if (name == triggerActionName && asObject["receiver"] != null){
                    return true;
                }
            }
            return false;
        }

        public JSONArray CollectionTriggerActions(string triggerID, string collectionName, string triggerActionType = StartTriggerActions)
        {
            var result = new JSONArray();
            var trigger = _c.containingAtom.GetStorableByID(triggerID) as CollisionTrigger;
            if (trigger == null) return null;
            JSONNode presentTriggers = trigger.trigger.GetJSON();
            var items = triggerActionType == StartTriggerActions ? presentTriggers["startActions"].AsArray : presentTriggers["endActions"].AsArray;
            var playClipActionName = $"{Storables.PlayClipFromCollectionAction}{collectionName}";
            var queueClipActionName = $"{Storables.QueueClipFromCollectionAction}{collectionName}";
            for (var i = 0; i < items.Count; i++) {
                var item = items[i].AsObject;
                if (item["receiver"] == null || item["receiverTargetName"] == null) continue;
                var receiverTarget = item["receiverTargetName"].ToString().Trim('"');
                if (receiverTarget == playClipActionName || receiverTarget == queueClipActionName) {
                    result.Add(item);
                }
            }

            return result;
        }

        public JSONClass AllCollectionTriggerActions(string collectionName)
        {
            var result = new JSONClass();
            var receiverName = _c.storeId;
            foreach (var triggerID in ColliderChoices)
            {
                var startTriggerActions = CollectionTriggerActions(triggerID, collectionName);
                var endTriggerActions = CollectionTriggerActions(triggerID, collectionName, EndTriggerActions);
                if (!startTriggerActions.Childs.Any() && !endTriggerActions.Childs.Any()) continue;
                result.Add(triggerID, new JSONClass()
                {
                    {"startActions", startTriggerActions},
                    {"endActions", endTriggerActions}
                });
            }
            return result;
        }

        public TriggerActionDiscrete AddTriggerAction(string triggerID, string triggerActionName, string receiverAtomID, string receiverNodeID, string triggerActionType = StartTriggerActions)
        {
            CleanUp();
            var trigger = _c.containingAtom.GetStorableByID(triggerID) as CollisionTrigger;
            if (trigger == null) return null;
            if (DoesTriggerActionExist(trigger, triggerActionName, triggerActionType)) return null;
            var newTriggerAction = triggerActionType == StartTriggerActions
                ? trigger.trigger.CreateDiscreteActionStartInternal()
                : trigger.trigger.CreateDiscreteActionEndInternal();
            newTriggerAction.name = triggerActionName;
            newTriggerAction.receiverAtom = _c.containingAtom;
            newTriggerAction.receiver = GetPluginStorableById(_c.GetContainingAtom(), receiverAtomID);
            newTriggerAction.receiverTargetName = receiverNodeID;
            trigger.enabled = true;

            return newTriggerAction;
        }

        public void RemoveTriggerAction(string triggerID, string triggerActionName, string triggerActionType, Action callback)
        {
            var trigger = _c.containingAtom.GetStorableByID(triggerID) as CollisionTrigger;
            if (trigger == null) return;
            var actionJSON = trigger.trigger.GetJSON();
            var index = 0;
            foreach (JSONNode triggerAction in actionJSON[triggerActionType].AsArray)
            {
                if (triggerAction["name"].ToString() == triggerActionName)
                {
                    actionJSON[triggerActionType][index] = null;
                    trigger.trigger.RestoreFromJSON(actionJSON);
                }
                index++;
            }

            callback();
        }

        public void UpdateTriggerAction(string triggerID, string oldTriggerActionName, string newTriggerActionName, string newTriggerActionReceiverName, string triggerActionType = StartTriggerActions)
        {
            var trigger = _c.containingAtom.GetStorableByID(triggerID) as CollisionTrigger;
            if (trigger == null) return;
            var actionJSON = trigger.trigger.GetJSON();
            var index = 0;

            foreach (JSONNode triggerAction in actionJSON[triggerActionType].AsArray)
            {
                if (triggerAction["name"].Value == oldTriggerActionName)
                {
                    actionJSON[triggerActionType][index]["name"] = newTriggerActionName;
                    actionJSON[triggerActionType][index]["receiverTargetName"] = newTriggerActionReceiverName;
                }
                index++;
            }

            trigger.trigger.RestoreFromJSON(actionJSON);
        }

        public JSONClass GetJSON(string collectionFilter = null)
        {
            var collections = new List<string>();
            if (collectionFilter == null)
            {
                collections.AddRange(_c.collections.Names);
            }
            else
            {
                collections.Add(collectionFilter);
            }

            var json = new JSONClass();
            foreach (var collectionName in collections)
            {
                var result = AllCollectionTriggerActions(collectionName);
                if (result.Childs.Any())
                {
                    foreach (KeyValuePair<string,JSONNode> trigger in result)
                    {
                        if (json[trigger.Key] == null)
                        {
                            json[trigger.Key] = new JSONClass()
                                {{"displayName", trigger.Key}};
                        }
                        if (trigger.Value["startActions"].Childs.Any())
                        {
                            foreach (var startAction in trigger.Value["startActions"].Childs)
                            {
                                startAction["receiverAtom"] = "Person";
                                startAction["receiver"] = "AudioMate";
                                json[trigger.Key]["startActions"][-1] = startAction;
                            }
                        }
                        if (trigger.Value["endActions"].Childs.Any())
                        {
                            foreach (var endAction in trigger.Value["endActions"].Childs)
                            {
                                endAction["receiverAtom"] = "Person";
                                endAction["receiver"] = "AudioMate";
                                json[trigger.Key]["endActions"][-1] = endAction;
                            }
                        }

                    }
                }
            }

            return json;
        }

        public void RemoveIncomingTriggerActions(string triggerID, string triggerActionType)
        {
            if (_c == null || _c.containingAtom == null) return;
            var trigger = _c.containingAtom.GetStorableByID(triggerID) as CollisionTrigger;
            if (trigger == null) return;
            var existing = trigger.trigger.GetJSON();
            if (!existing.Childs.Any() || !existing[triggerActionType].Childs.Any()) return;
            var receiverAtom = _c.containingAtom.uid;
            var receiver = _c.storeId;
            var index = 0;
            foreach (var tA in existing[triggerActionType].Childs)
            {
                if (tA == null || tA["receiverAtom"] == null || tA["receiver"] == null) continue;
                if (tA["receiverAtom"].Value == receiverAtom &&
                    tA["receiver"].Value == receiver)
                {
                    existing[triggerActionType][index] = new JSONNode();
                }
                index++;
            }
            trigger.trigger.RestoreFromJSON(existing);
        }

        public void RestoreFromJSON(JSONNode json)
        {
            if (_c == null || json == null || !json.Childs.Any()) return;
            var receiverAtom = _c.containingAtom.uid;
            var receiver = _c.storeId;
            foreach (var triggerActions in json.Childs)
            {
                var triggerID = triggerActions["displayName"];
                if (triggerID == null) continue;
                var trigger = _c.containingAtom.GetStorableByID(triggerID) as CollisionTrigger;
                if (trigger == null) continue;
                if (triggerActions["startActions"].Childs.Any())
                {
                    var index = 0;
                    RemoveIncomingTriggerActions(triggerID, "startActions");
                    foreach (var tA in triggerActions["startActions"].Childs)
                    {
                        tA["receiverAtom"] = receiverAtom;
                        tA["receiver"] = receiver;
                        triggerActions["startActions"][index] = tA;
                        index++;
                    }
                }
                if (triggerActions["endActions"].Childs.Any())
                {
                    var index = 0;
                    RemoveIncomingTriggerActions(triggerID, "endActions");
                    foreach (var tA in triggerActions["endActions"].Childs)
                    {
                        tA["receiverAtom"] = receiverAtom;
                        tA["receiver"] = receiver;
                        triggerActions["endActions"][index] = tA;
                        index++;
                    }
                }
                trigger.trigger.RestoreFromJSON(triggerActions.AsObject, null, true);
            }
        }

        public void OnDestroy()
        {
            // Nothing to see here...
        }

        public void OnAtomRename(string oldID, string newID)
        {
            // Nothing to see here...
        }

        private void Log(string message)
        {
            if (_c == null) return;
            _c.Log($"TriggerManager: {message}");
        }
    }
}
