﻿using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

namespace AudioMate
{

    public class UIClipLibrary : UIComponent
    {
        private static ScrollRect _content;

        public static UIClipLibrary AddTo(Transform parent)
        {
            var go = new GameObject();
            go.transform.SetParent(parent, false);
            _content = UIFactory.CreateScrollRect(go.gameObject);
            var layout = go.AddComponent<LayoutElement>();
            layout.flexibleHeight = 1;
            layout.preferredHeight = 875f;
            return Configure(go);
        }

        private static UIClipLibrary Configure(GameObject go)
        {
            var group = go.AddComponent<VerticalLayoutGroup>();
            group.spacing = 5f;

            var rect = go.GetComponent<RectTransform>() ?? go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1);
            rect.StretchBottom();

            return go.AddComponent<UIClipLibrary>();
        }

        private static GameObject CreateContainer()
        {
            var go = new GameObject("ClipLibrary Container");
            go.transform.SetParent(_content.content, false);
            var layout = go.AddComponent<LayoutMaxSize>();
            layout.adjustHeight = true;
            layout.adjustWidth = true;
            layout.SetLayoutVertical();
            {
                var child = new GameObject();
                child.transform.SetParent(go.transform, false);

                var rect = child.AddComponent<RectTransform>();
                rect.StretchParent();
            }

            return go;
        }

        private List<ClipUI> _drawnElements = new List<ClipUI>();

        public UIFactory prefabFactory;

        public bool isRefreshing;

        public bool isFilterOn;

        protected override void _Init()
        {
            var container = CreateContainer();
            prefabFactory = container.gameObject.AddComponent<UIFactory>();
            prefabFactory.c = _c;
            RefreshAll();
            _c.clipLibrary.OnClipsRefreshed.AddListener(OnClipsRefreshed);
        }

        public void OnClipsRefreshed()
        {
            RefreshAll();
        }

        public void RefreshAll()
        {
            if (isRefreshing || _c == null || _c.Clips == null) return;
            isRefreshing = true;
            Log("Refreshing Clip Library UI");
            try
            {
                // DestroyClipUI();
                RefreshAllClipUIs();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(UIClipLibrary)}.{nameof(RefreshAll)}: {e}");
            }
            finally
            {
                isRefreshing = false;
            }
        }

        public void RefreshAllClipUIs()
        {
            Log("Refreshing Clip Library UI");
            if (_drawnElements == null)
            {
                Log("UI container is not initialised. Aborting BuildClipUI.");
                return;
            }
            if (_c.Clips == null || _c.Clips.Count <= 0)
            {
                Log("No Clips found. Aborting BuildClipUI.");
                return;
            }
            try
            {
                var activeCollection = _c.collections.ActiveCollection;
                var destroyedClips = new List<AudioMateClip>();
                var provisionalClipsCount = _c.Clips.FindAll(x => x.IsProvisional).Count;
                Log($"Drawn elements: {_drawnElements.Count}, Clips: {_c.Clips.Count}, Provisional Clips: {provisionalClipsCount}, Active Collection Count: {activeCollection.Count}");
                foreach (var clip in _c.Clips)
                {
                    if (clip.AudioClip == null || clip.AudioClip?.destroyed == true)
                    {
                        Log($"Destroying removed Clip {clip.UID}");
                        destroyedClips.Add(clip);
                        clip.Destroy();
                        continue;
                    }
                    if (isFilterOn)
                    {
                        if (!activeCollection.Contains(clip))
                        {
                            DestroyClipUI(clip);
                            continue;
                        }
                    }
                    // Only create a new Clip UI if it doesn't exist, isn't initialised or hasn't been added to the
                    // tab view.
                    if (clip.UI == null || clip.UI.IsInitialised == false || !_drawnElements.Contains(clip.UI))
                    {
                        var clipUI = CreateClipUI(clip);
                        _drawnElements.Add(clipUI);
                    }

                    clip.SetState(activeCollection.Contains(clip));
                }
                // Remove destroyed Clips from Clip Library
                foreach (var clip in destroyedClips)
                {
                    _c.Clips.Remove(clip);
                }
            }
            catch (Exception e)
            {
                Log($"{e})");
            }

        }

        public void DestroyClipUI(AudioMateClip clip)
        {
            _drawnElements.Remove(clip.UI);
            clip.DestroyUI();
        }

        public void SetFilter(bool isOn)
        {
            isFilterOn = isOn;
            RefreshAllClipUIs();
        }

        // Converts a NamedAudioClip into a AudioMateClip and creates the necessary UI to display the clip
        // in the clip library
        private ClipUI CreateClipUI(AudioMateClip clip)
        {
            var go = new GameObject(clip.UID);
            go.transform.SetParent(_content.content, false);
            var gridLayout = go.AddComponent<HorizontalLayoutGroup>();
            gridLayout.spacing = 10f;
            gridLayout.childForceExpandWidth = false;
            gridLayout.childControlWidth = true;
            gridLayout.childForceExpandHeight = true;
            gridLayout.childControlHeight = true;
            gridLayout.padding = new RectOffset(0, 5, 0, 0);
            var buttonGroup = gridLayout.gameObject.AddComponent<HorizontalButtonGroup>();
            var previewButton = buttonGroup.CreateButton(Symbols.Play, Styles.Success, () =>
            {
                clip.AudioClip?.Test();
                _c.Cursor.Set(clip);
            });
            var toggleButton = buttonGroup.CreateButton(clip.DisplayName, Styles.Disabled, () =>
            {
                clip.ToggleState();
                if (clip.IsInActiveCollection)
                {
                    _c.collections.AddClipToActiveCollection(clip);
                }
                else
                {
                    _c.collections.RemoveClipFromActiveCollection(clip);
                }
                _c.Cursor.Set(clip);
            }, true, 0, 100f);
            clip.UI.Container = go.transform;
            clip.UI.PreviewButton = previewButton;
            clip.UI.ToggleButton = toggleButton;
            clip.UI.IsInitialised = true;
            clip.RefreshUI();
            return clip.UI;
        }

        public override void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs eventArgs)
        {
            Log("A new active Collection has been selected");
            RefreshAllClipUIs();
        }

        public override void OnActiveCollectionUpdated()
        {
            Log("Active Collection has been updated");
            RefreshAllClipUIs();
        }

        protected override void Log(string message)
        {
            if (_c == null) return;
            _c.Log($" [UIClipLibrary]: {message}");
        }
    }
}
