﻿using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine.UI;

namespace AudioMate
{
    public class ScreenCollection : Screen
    {
        public new static readonly ScreenDefinition Definition = new ScreenDefinition(
            "Collection",
            "Collection",
            typeof(ScreenCollection));

        private JSONStorableBool _toggleEnabledJSON;
        private JSONStorableFloat _sliderVolumeJSON;
        private JSONStorableFloat _sliderPitchJSON;
        private JSONStorableFloat _sliderPlayChanceJSON;
        private JSONStorableBool _toggleOnlyIfClearJSON;
        private JSONStorableBool _toggleOnlyIfThisCollectionIsNotPlayingJSON;
        private JSONStorableBool _toggleAlwaysQueueJSON;
        private JSONStorableBool _toggleShuffleJSON;
        private JSONStorableBool _toggleAutoPlayJSON;
        private JSONStorableBool _toggleContinuousPlayJSON;
        private Text _textHeading;

        private bool _needsRefresh;

        private readonly List<HorizontalButtonGroup> _triggerActionsUI = new List<HorizontalButtonGroup>();

        public override void Init(AudioMateController controller)
        {
            base.Init(controller);
            InitPlaybackOptionsUI();
            InitTriggerActionsUI();
        }

        private void InitTriggerActionsUI()
        {
            _ui.CreateHeader("Trigger Actions", 2);
            if (_c.DisplayHints) {
                _ui.CreateText("Add and remove bindings to physical trigger actions here. If a binding is active, it will play this collection when the selected trigger is fired.", FontSizes.Small);
                _ui.CreateText("'Start Actions' will run as soon as the trigger is fired, 'End Actions' will run when the trigger has stopped.", FontSizes.Small);
            }
            _ui.CreatePopup(_c.TriggerColliderChooserJSON, true, 250f);
            _c.TriggerColliderChooserJSON.setCallbackFunction = val =>
            {
                _c.TriggerColliderChooserJSON.choices = _c.Triggers.ColliderChoices;
                _needsRefresh = true;
            };

            var btnGroup = _ui.gameObject.AddComponent<HorizontalButtonGroup>();
            btnGroup.CreateButton("Add Start Action", Styles.Button, () =>
            {
                _c.OnAddTriggerButtonClicked();
                _needsRefresh = true;
            });
            btnGroup.CreateButton("Add End Action", Styles.Button, () =>
            {
                _c.OnAddTriggerButtonClicked(TriggerManager.EndTriggerActions);
                _needsRefresh = true;
            });

            _needsRefresh = true;
        }

        private void InitPlaybackOptionsUI()
        {
            _textHeading = _ui.CreateHeader("Untitled Settings", 2);
            if (_c.DisplayHints) {
                _ui.CreateText("These are the settings for the currently active (selected) collection. Use the drop-down and the buttons on the right side to select, create, clear or delete a collection.", FontSizes.Small);
                _ui.CreateText("To change the name of this collection, click on the big white name on the right side and enter a new one.", FontSizes.Small);
            }
            _toggleEnabledJSON = new JSONStorableBool("Enabled", true, (bool val) =>
            {
                _c.collections.ActiveCollection.Enabled = val;
            });
            _ui.CreateToggle(_toggleEnabledJSON);
            _ui.CreateSpacer(3f);
            if (_c.VolumeControlEnabledJSON.val)
            {
                _sliderVolumeJSON = new JSONStorableFloat("Volume", 1f,
                    (float val) => { _c.collections.ActiveCollection.Volume = val; }, 0f, 1f);
                _ui.CreateSlider(_sliderVolumeJSON, 1F, true);
                _ui.CreateSpacer(3f);
            }
            if (_c.PitchControlEnabledJSON.val)
            {
                _sliderPitchJSON = new JSONStorableFloat("Pitch", 1f,
                    (float val) => { _c.collections.ActiveCollection.Pitch = val; }, 0f, 2f);
                _ui.CreateSlider(_sliderPitchJSON, 1f, true);
                _ui.CreateSpacer(3f);
            }
            _sliderPlayChanceJSON = new JSONStorableFloat("Play Chance", 1f, (float val) =>
            {
                _c.collections.ActiveCollection.PlayChance = val;
            }, 0f, 1f);
            _ui.CreateSlider(_sliderPlayChanceJSON, 1f, true);
            _ui.CreateSpacer(3f);

            _toggleOnlyIfClearJSON = new JSONStorableBool("Play Only If Audio Source Is Clear", false, (bool val) =>
            {
                _c.collections.ActiveCollection.OnlyIfClear = val;
                if (val)
                {
                    _toggleOnlyIfThisCollectionIsNotPlayingJSON.val = false;
                }
            });
            _ui.CreateToggle(_toggleOnlyIfClearJSON);

            _toggleOnlyIfThisCollectionIsNotPlayingJSON = new JSONStorableBool("Play Only If This Collection Isn't Playing", false, (bool val) =>
            {
                _c.collections.ActiveCollection.OnlyIfThisCollectionIsNotPlaying = val;
                if (val)
                {
                    _toggleOnlyIfClearJSON.val = false;
                }
            });
            _ui.CreateToggle(_toggleOnlyIfThisCollectionIsNotPlayingJSON);

            _toggleAlwaysQueueJSON = new JSONStorableBool("Always Queue", false, (bool val) =>
            {
                _c.collections.ActiveCollection.AlwaysQueue = val;
            });
            _ui.CreateToggle(_toggleAlwaysQueueJSON);

            _toggleShuffleJSON = new JSONStorableBool("Shuffle Mode", true, (bool val) =>
            {
                _c.collections.ActiveCollection.ShuffleMode = val;
            });
            _ui.CreateToggle(_toggleShuffleJSON);

            _toggleAutoPlayJSON = new JSONStorableBool("Auto Play", false, (bool val) =>
            {
                _c.collections.ActiveCollection.HasAutoPlayBeenTriggered = false;
                _c.collections.ActiveCollection.AutoPlay = val;
            });
            _ui.CreateToggle(_toggleAutoPlayJSON);

            _toggleContinuousPlayJSON = new JSONStorableBool("Continuous Play", false, (bool val) =>
            {
                _c.collections.ActiveCollection.ContinuousPlay = val;
            });
            _ui.CreateToggle(_toggleContinuousPlayJSON);

            var activeCollection = _c.collections.ActiveCollection;
            InitSettings(activeCollection);
        }

        private void InitSettings(Collection activeCollection)
        {
            if (activeCollection == null) return;
            _textHeading.text = $"{activeCollection.Name} Settings";
            _toggleEnabledJSON.val = activeCollection.Enabled;
            _toggleOnlyIfClearJSON.val = activeCollection.OnlyIfClear;
            _toggleOnlyIfThisCollectionIsNotPlayingJSON.val = activeCollection.OnlyIfThisCollectionIsNotPlaying;
            _toggleAlwaysQueueJSON.val = activeCollection.AlwaysQueue;
            _toggleShuffleJSON.val = activeCollection.ShuffleMode;
            _toggleAutoPlayJSON.val = activeCollection.AutoPlay;
            _toggleContinuousPlayJSON.val = activeCollection.ContinuousPlay;
            if (_c.VolumeControlEnabledJSON.val) _sliderVolumeJSON.val = activeCollection.Volume;
            if (_c.PitchControlEnabledJSON.val) _sliderPitchJSON.val = activeCollection.Pitch;
            _sliderPlayChanceJSON.val = activeCollection.PlayChance;
        }

        private void RefreshTriggerActionsUI()
        {
            _c.TriggerColliderChooserJSON.choices = _c.Triggers.ColliderChoices;
            _c.TriggerColliderChooserJSON.displayChoices = _c.Triggers.ColliderChoices;

            var activeCollection = _c.collections.ActiveCollection;
            var triggerCollection = _c.Triggers.AllCollectionTriggerActions(activeCollection.Name);
            foreach (var uiElem in _triggerActionsUI)
            {
                uiElem.Destroy();
            }

            var triggerID = _c.TriggerColliderChooserJSON.val;
            var triggerActions = triggerCollection[triggerID];
            if (triggerActions[TriggerManager.StartTriggerActions].Childs.Any()) {
                foreach (JSONClass tA in triggerActions[TriggerManager.StartTriggerActions].AsArray)
                {
                    CreateTriggerActionUI(triggerID, tA, TriggerManager.StartTriggerActions);
                }
            }
            if (triggerActions[TriggerManager.EndTriggerActions].Childs.Any()) {
                foreach (JSONClass tA in triggerActions[TriggerManager.EndTriggerActions].AsArray)
                {
                    CreateTriggerActionUI(triggerID, tA, TriggerManager.EndTriggerActions);
                }
            }
        }

        private void CreateTriggerActionUI(string triggerID, JSONNode triggerActionJSON, string triggerActionType)
        {
            var buttonGroup = _ui.CreateButtonGroup();
            buttonGroup.CreateLabel(triggerActionType == TriggerManager.StartTriggerActions ? "Start Trigger Action" : "End Trigger Action", Styles.Button, true, 26);
            buttonGroup.CreateButton(Symbols.Remove, Styles.Danger, () =>
            {
                _needsRefresh = true;
                _c.Triggers.RemoveTriggerAction(triggerID, triggerActionJSON["name"].ToString(), triggerActionType,
                    () => RefreshTriggerActionsUI());
            }, false, 50f, 0);

            _triggerActionsUI.Add(buttonGroup);
        }

        protected override void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
            base.OnActiveCollectionSelected(args);
            var activeCollection = args.After;
            InitSettings(activeCollection);
            _needsRefresh = true;
        }

        private void FixedUpdate()
        {
            if (!_needsRefresh) return;
            _needsRefresh = false;
            RefreshTriggerActionsUI();
        }
    }
}
