﻿using System;
using System.Collections;
using UnityEngine;
using UnityEngine.UI;

namespace AudioMate
{
    [RequireComponent(typeof(UIDynamicButton))]
    public class UIManager : MonoBehaviour
    {
        public static UIManager AddTo(Transform transform)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);

            var topLevelFactory = go.AddComponent<UIFactory>(); ;


            var tabs = Tabs.AddTo(go.transform);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchTop();
            rect.pivot = new Vector2(0, 1);

            go.AddComponent<VerticalLayoutGroup>();

            var panels = new GameObject();
            panels.transform.SetParent(rect.transform, false);

            var panelsGroup = panels.AddComponent<HorizontalLayoutGroup>();
            panelsGroup.spacing = 10f;
            panelsGroup.childControlWidth = true;
            panelsGroup.childForceExpandWidth = false;
            panelsGroup.childControlHeight = true;
            panelsGroup.childForceExpandHeight = true;

            var leftPanel = CreatePanel(panels.transform, 0f, 1f);
            var rightPanel = CreatePanel(panels.transform, 450f, 0f, true);

            var ui = go.AddComponent<UIManager>();
            ui.container = topLevelFactory;
            ui.tabs = tabs;
            ui.leftPanel = leftPanel;
            ui.rightPanel = rightPanel;
            return ui;
        }

        private static GameObject CreatePanel(Transform transform, float preferredWidth, float flexibleWidth, bool embedded = false)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1f);
            var layout = go.AddComponent<LayoutElement>();
            layout.minWidth = 0;
            layout.preferredWidth = preferredWidth;
            layout.flexibleWidth = flexibleWidth;
            layout.flexibleHeight = 1;

            if (embedded)
            {
                var image = go.gameObject.AddComponent<Image>();
                image.color = Styles.EmbeddedBg;
                image.preserveAspect = false;
                var verticalLayoutGroup = go.AddComponent<VerticalLayoutGroup>();
                verticalLayoutGroup.childControlWidth = true;
                verticalLayoutGroup.childForceExpandWidth = true;
                verticalLayoutGroup.childControlHeight = true;
                verticalLayoutGroup.childForceExpandHeight = false;
                verticalLayoutGroup.spacing = 10f;
                verticalLayoutGroup.padding = new RectOffset(20, 20, 20, 20);
            }
            else
            {
                var verticalLayoutGroup = go.AddComponent<VerticalLayoutGroup>();
                verticalLayoutGroup.childControlWidth = true;
                verticalLayoutGroup.childForceExpandWidth = true;
                verticalLayoutGroup.childControlHeight = true;
                verticalLayoutGroup.childForceExpandHeight = false;
                verticalLayoutGroup.spacing = 5f;
            }



            return go;
        }
        public ScreenManager screenManager;

        private AudioMateController _c;

        public UIFactory container;
        public Tabs tabs;
        public GameObject leftPanel;
        public GameObject rightPanel;
        public UIActiveCollection activeCollectionUI;
        private OverlayTemplates _oTemplates = new OverlayTemplates();


        #region Init
        public void Init(AudioMateController mainController)
        {
            Log("Initializing UI Manager");
            if (mainController == null)
            {
                SuperController.LogError("AudioMate: Error during UI initialization. Controller or Collection Manager are set to null.");
                return;
            }
            _c = mainController;
            container.c = _c;
            Log("Initializing Screen Manager");
            screenManager = InitScreenManager(leftPanel);
            Log("Initializing Active Collection UI#");
            InitActiveCollectionUI();
            _c.OnUIInitialized();
        }

        private static ScreenManager InitScreenManager(GameObject panel)
        {
            var go = new GameObject();
            go.transform.SetParent(panel.transform, false);

            var layout = go.AddComponent<LayoutElement>();
            layout.flexibleHeight = 1;
            layout.flexibleWidth = 1;

            var screensManager = ScreenManager.AddTo(go);
            return screensManager;
        }

        public void InitActiveCollectionUI()
        {
            if (_c == null) return;
            activeCollectionUI = UIActiveCollection.AddTo(rightPanel.transform);
            activeCollectionUI.Init(_c);
        }

        public void Bind()
        {
            if (_c == null)
            {
                SuperController.LogError("AudioMate: Error during bind. Can't access main controller.");
                return;
            }
            try
            {
                Log("Binding UI Manager");

                // Screens are defined in ScreenManager.cs
                Screens.List.ForEach(screen => tabs.Add(screen.ID, screen.Title));

                screenManager.OnScreenChanged.AddListener(screenID => tabs.Select(screenID));
                screenManager.Bind(_c);

                tabs.OnTabSelected.AddListener(screenManager.ChangeScreen);

                activeCollectionUI.Bind();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(UIManager)}.{nameof(Bind)}: {e}");
            }
        }
        #endregion

        public void ShowOkOverlay(string message, Action callback = null)
        {
            _oTemplates.Ok.Callback = callback;
            container.CreateOverlay(message, _oTemplates.Ok);
        }

        public void ShowYesNoOverlay(string message, Action yesCallback = null, Action noCallback = null)
        {
            var template = _oTemplates.YesNo;
            template.MainButtons[0].Callback = yesCallback;
            template.MainButtons[1].Callback = noCallback;
            container.CreateOverlay(message, template);
        }

        #region Callbacks
        public void OnDestroy()
        {
            if (rightPanel != null) Destroy(rightPanel);
            if (leftPanel != null) Destroy(leftPanel);
        }
        #endregion

        #region Helpers
        private void Log(string message)
        {
            if ((UnityEngine.Object) _c == (UnityEngine.Object) null) return;
            _c.Log($" [UIManager]: {message}");
        }
        #endregion
    }
}
